
indexing
	description: "Lightweight node for data centric XML"
	description_datacentric: "data centric means no mixed content, that is%
		%elements contain EITHER (significant)text OR child elements"
	license: "http://www.opensource.org/licenses/mit-license.html"
	revision: "$Id: nxml_element.e,v 1.6 2001/07/14 00:52:50 farnaud Exp $"
	-- TODO:
	-- as_xml vs. as_xml_with map
	-- escape text and attribute content in as_xml
	-- empty element
	-- finish is_namespace_map
	-- is_equal
	
class NXML_ELEMENT

inherit
	ANY
		redefine
			out
		end
		
creation
	make_root,
	make_child
	
feature {NONE} -- Creation

	make_root (a_name: NXML_NAME)is
			-- Make root element.
		require
			not_void_name: a_name /= Void
		do
			name := a_name
			make_elements
			make_attributes
			make_text
		ensure
			root: is_root
			starts_as_leaf: is_leaf
		end

	make_child (a_parent: NXML_ELEMENT; a_name: NXML_NAME) is
			-- Make with name and no content or attributes .
		require
			not_void_parent: a_parent /= Void
			not_void_name: a_name /= Void
		do
			make_root (a_name)
			parent := a_parent
		ensure
			not_root: not is_root
			starts_as_leaf: is_leaf
		end
		
feature -- Data

	parent: NXML_ELEMENT
			-- Parent
			-- require not is_root
			-- ensure Result /= Void
			
	name: NXML_NAME
			-- Name.
	
	attributes: NXML_NAME_TABLE [STRING]
			-- Attributes
	
	elements: DS_BILINKED_LIST [NXML_ELEMENT]
			-- Elements which are under this element.

feature {NONE} -- Data

	text_impl: STRING
	keyed_elements_impl: NXML_NAME_TABLE [NXML_ELEMENT]
	
feature -- Access

	is_root: BOOLEAN is
			-- Is this the root node?
		do
			Result := parent = Void
		ensure
			definition: Result = (parent = Void)
		end

	depth: INTEGER is
			-- Depth from root.
		local
			a_node: NXML_ELEMENT
		do
			-- recursion in a loop
			from
				Result := 1
				a_node := Current
			until
				a_node.is_root
			loop
				a_node := a_node.parent
				Result := Result + 1
			end
		ensure
			start: is_root implies (Result = 1)
			recurse: is_root or else (Result = (parent.depth + 1))
		end
		
	is_leaf: BOOLEAN is
			-- Is this a final node with no child elements?
		do
			Result := elements.count = 0
		end
			
	text: STRING is
			-- Text content of leaf or markup (post parsing) of non-leaf.
			-- Mixed content's text, comments etc are ignored.
		local
			a_cursor: DS_LINEAR_CURSOR [NXML_ELEMENT]
		do
			if is_leaf then
				Result := text_impl
			else
				!! Result.make (0)
				from
					a_cursor := elements.new_cursor
					a_cursor.start
				until
					a_cursor.after
				loop
					Result.append_string (a_cursor.item.out)
					a_cursor.forth
				end
			end
		ensure
			not_void: Result /= Void
		end

	namespaces: DS_SPARSE_SET [NXML_NAMESPACE] is
			-- All namespaces used in this element and child elements.
		local
			attr_it: DS_SPARSE_TABLE_CURSOR [STRING,NXML_NAME]
			it: DS_LINEAR_CURSOR[NXML_ELEMENT]
		do
			-- create set based on is_equal comparison
			!DS_HASH_SET [NXML_NAMESPACE]! Result.make_equal (0)

			-- add element's namespace
			Result.force_new (name.namespace)
			
			-- add element
			from
				attr_it := attributes.new_cursor
				attr_it.start
			until
				attr_it.after
			loop
				Result.force (attr_it.key.namespace)
				attr_it.forth
			end
			
			-- union with child namespaces
			from
				it := elements.new_cursor
				it.start
			until
				it.after
			loop
				Result.append (it.item.namespaces)
				it.forth
			end
		end

	keyed_elements: NXML_NAME_TABLE [NXML_ELEMENT] is
			-- Elements keyed by name. Only the first instance of 
			-- two or more element with the same name is in this 
			-- table.
			-- Warning: generated once per instance.
		local
			it: DS_LINEAR_CURSOR[NXML_ELEMENT]
		do
			if keyed_elements_impl = Void then	
				-- initialise keyed elements
				!! keyed_elements_impl.make (name.namespace)
				from
					it := elements.new_cursor
					it.start
				until
					it.after
				loop
					if not keyed_elements_impl.has (it.item.name) then
						keyed_elements_impl.force_new (it.item, it.item.name)
					end
					it.forth
				end
			end
			Result := keyed_elements_impl
		end

feature -- Text output

	out: STRING is
			-- Pseudo XML representation of element.
			-- (pseudo in that namespaces are resolved and thus invalid)
		local
			it: DS_SPARSE_TABLE_CURSOR [STRING, NXML_NAME]
		do
			!! Result.make (0)
			Result.append_string ("<")
			Result.append_string (name.out) -- non xml!
			from
				it := attributes.new_cursor
				it.start
			until
				it.after
			loop
				Result.append_character (' ')
				Result.append_string (it.key.out) -- non xml!
				Result.append_character ('=')
				Result.append_character ('%'')
				Result.append_string (it.item)
				Result.append_character ('%'')
				it.forth
			end
			Result.append_string (">")

			-- content
			Result.append_string (text)

			Result.append_string ("</")
			Result.append_string (name.out) -- non xml!
			Result.append_string (">")
		end
		
	as_xml (a_ns_map: DS_SPARSE_TABLE [STRING, NXML_NAMESPACE]): STRING is
			-- XML representation, using a namepace prefix resolution table.
		require
			ns_map: is_namespace_map (a_ns_map)
		do
			!! Result.make (0)
			as_xml_in (Result, a_ns_map)
		end
			
	as_xml_in (a_string:STRING; a_ns_map: DS_SPARSE_TABLE [STRING, NXML_NAMESPACE]) is
			-- Write as_xml in string.
		require
			appendable: a_string /= Void
			ns_map: is_namespace_map (a_ns_map)
		local
			it: DS_LINEAR_CURSOR [NXML_ELEMENT]
		do
			append_depth_in (a_string)
			a_string.append_character ('<')
			
			-- element name and its namespace declaration if needed
			a_string.append_string (name.name)
			if not is_parent_namespace (name.namespace) then
				a_string.append_character (' ')
				name.namespace.namespace_declaration_in (a_string)
			end
			
			-- all top level xml ns declarations
			if is_root then
				-- write ns declarations
			end
			
			-- write attributes
			append_attributes_in (a_string, a_ns_map)
			
			-- remainder
			if is_leaf and text.count = 0 then
				-- empty element
				a_string.append_character ('/')
				a_string.append_character ('>')
			elseif is_leaf then
				-- terminate start tag
				a_string.append_character ('>')
				
				-- text
				a_string.append_string (text) -- ESCAPE!
				append_end_tag_in (a_string)
			else
				-- finish tag
				a_string.append_character ('>')
				
				-- start tag indentation
				append_new_line_in (a_string)
				
				-- sub elements
				from
					it := elements.new_cursor
					it.start
				until
					it.after
				loop
					-- recurse
					it.item.as_xml_in (a_string, a_ns_map)
					it.forth
				end
				
				-- end tag indentation
				append_depth_in (a_string)
				append_end_tag_in (a_string)
			end
			
			append_new_line_in (a_string)
		end

	is_namespace_map (a_ns_map: DS_SPARSE_TABLE [STRING, NXML_NAMESPACE]): BOOLEAN is
			-- Is this a complete namespace map for this element?
		do
			Result := a_ns_map /= Void --... to complete
		end

	default_namespaces_map: DS_HASH_TABLE [STRING, NXML_NAMESPACE] is
			-- Generate names for namespaces (ns123).
		local
			it: DS_SPARSE_SET_CURSOR [NXML_NAMESPACE]
			a_prefix: STRING
			i: INTEGER
		do
			!! Result.make_default
			from
				it := namespaces.new_cursor
				it.start
				i := 1
			until
				it.after
			loop
				!! a_prefix.make (0)
				a_prefix.append_string (Ns_map_prefix)
				a_prefix.append_string (i.out)
				Result.force (a_prefix, it.item)
				
				i := i + 1
				it.forth
			end
		end

feature {NONE} -- Text output support
				
	is_parent_namespace (a_namespace: NXML_NAMESPACE): BOOLEAN is
			-- Is this namespace the same as the parent's?
		require
			not_void: a_namespace /= Void
		do
			Result :=  (not is_root) 
				and then parent.name.namespace.is_equal (a_namespace)
		ensure
			def: Result = ((not is_root) and then parent.name.namespace.is_equal (a_namespace))
		end

	append_declarations_in (a_string: STRING; a_ns_map: DS_SPARSE_TABLE [STRING, NXML_NAMESPACE]) is
			-- Append start tag attribute declarations in string.
		require
			appendable: a_string /= Void
			ns_map: is_namespace_map (a_ns_map)
		local
			it: DS_SPARSE_TABLE_CURSOR [STRING, NXML_NAMESPACE]
		do
			from
				it := a_ns_map.new_cursor
				it.start
			until
				it.after
			loop
				a_string.append_character (' ')
				it.key.namespace_prefix_declaration_in (a_string, it.item)
				it.forth
			end
		end

	append_attributes_in (a_string: STRING; a_ns_map: DS_SPARSE_TABLE [STRING, NXML_NAMESPACE]) is
			-- Append start tag attribute declarations in string.
		require
			appendable: a_string /= Void
			ns_map: is_namespace_map (a_ns_map)
		local
			attr_it: DS_SPARSE_TABLE_CURSOR [STRING, NXML_NAME]
		do
			from
				attr_it := attributes.new_cursor
				attr_it.start
			until
				attr_it.after
			loop
				a_string.append_character (' ')
				
				-- no prefix if same prefix as element name
				if name.same_namespace (attr_it.key) then
					a_string.append_string (attr_it.key.name)
				else
					attr_it.key.prefixed_in (a_string, a_ns_map)
				end
				a_string.append_character ('=')
				a_string.append_character ('%'')
				a_string.append_string (attr_it.item) -- ESCAPE!
				a_string.append_character ('%'')

				attr_it.forth
			end
		end
			
	append_end_tag_in (a_string: STRING) is
			-- Append end tag.
		require
			appendable: a_string /= Void
		do
			a_string.append_character ('<')
			a_string.append_character ('/')
			a_string.append_string (name.name)
			a_string.append_character ('>')
		end
	
	append_depth_in (a_string: STRING) is
			-- Indentation: append spaces x depth.
		require
			appendable: a_string /= Void
		local
			i: INTEGER
		do
			if do_indent then
				from
					i := depth
				until
					i = 1
				loop
					a_string.append_character (' ')
					i := i - 1
				end
			end
		ensure
			count: a_string.count = (old a_string.count) + depth - 1
		end
		
	append_new_line_in (a_string: STRING) is
			-- Indentation: append end of line.
		require
			appendable: a_string /= Void
		do
			if do_indent then
				a_string.append_character ('%N')
			end
		end
		
	do_indent: BOOLEAN is
			-- Indentation flag.
		do
			Result := True
		end

	Ns_map_prefix: STRING is "ns"
	
feature {NONE} -- Creation support

	make_attributes is
			-- Initialise attributes.
			-- The name must be set and constant because the attributes 
			-- table shares the namespace.
		do
			!! attributes.make (name.namespace)
		ensure
			not_void: attributes /= Void
			empty: attributes.count = 0
		end
		
feature {NONE} -- Creation support

	make_text is
			-- Default text (shared empty string).
		do
			text_impl := shared_empty_string
		ensure
			not_void: text_impl /= Void
		end
		
	shared_empty_string: STRING is
			-- Default empty text (shared).
		once
			!! Result.make (0)
		ensure
			not_void: Result /= Void
			count: Result.count = 0
		end
		
feature {NONE} -- Creation support
	
	make_elements is
			-- This allocates a unique empty list so that leaf nodes
			-- do not create extra data structure. `add_child' replaces 
			-- this list on the first child addition.
		do
			elements := shared_empty_list
		end
		
	shared_empty_list: DS_BILINKED_LIST[NXML_ELEMENT] is
			-- Empty list shared by all instances.
		once
			!! Result.make
		end
	
feature {NXML_ANY} -- Construction

	add_element (a_child: NXML_ELEMENT) is
			-- Add a child (copes with shared empty list).
		do
			if elements = shared_empty_list then
				!! elements.make
			end
			elements.force_last (a_child)
			
			-- reset keyed elements cache
			keyed_elements_impl := Void
		ensure
			count: elements.count = ((old (elements.count)) + 1)
		end

	set_text (a_text: STRING) is
			-- Set text.
		require
			not_void: a_text /= Void
			leaf: is_leaf
		do
			text_impl := a_text
		ensure
			set: text.is_equal (a_text)
		end
				
invariant
	root: is_root = (parent = Void)
	name_not_void: name /= Void
	elements_not_void: elements /= Void

end
