indexing
	description: "Parser that produces a stream of event with namespaces resolved"
	project: "http://www.nenie.org/eiffel/"
	license: "http://www.opensource.org/licenses/mit-license.html"
	revision: "$Id: nxml_parser_namespace.e,v 1.5 2001/07/13 19:59:13 farnaud Exp $"
	-- TODO
	-- check duplicate namespace declarations
		
deferred class
	NXML_PARSER_NAMESPACE

inherit
	NXML_NAMESPACE_EVENT_INTERFACE
	NXML_PARSER
	NXML_ANY

feature -- Actions

	when_start_tag (a_name: NXML_PARSER_NAME) is
			-- Start tag.
		do
			-- init namespaces stack if root element
			if namespaces = Void then
				make_namespaces
				make_all_names
			end
			
			-- init element state
			make_start_tag (a_name)
		end
	
	
	when_start_tag_attribute (a_name: NXML_PARSER_NAME; a_value: STRING) is		
			-- gather namespaces first, as they can be used in the 
			-- element name or in tags of sibling attributes placed
			-- before.
		local
			a_namespace: NXML_NAMESPACE
		do
			if a_name.is_namespace_declaration then
				-- build context if not already used, it 
				-- inherits parents namespaces which will be 
				-- overridden by those declared here
				if start_tag_context = Void then
					!! start_tag_context.make
					start_tag_context.copy (namespace_context)
				end

				-- bind namespace
				!! a_namespace.make (a_value)
				if a_name.is_named_namespace_declaration then
					-- named namespace
					start_tag_context.add_named (a_namespace,a_name.last)
				else 
					-- default namespace declaration
					start_tag_context.set_namespace (a_namespace)
				end
			else
				attribute_names.force (a_name)
				attribute_values.force (a_value)
			end
		end
		
	when_start_tag_end is
			-- End of tag, process resolved start tag names.
		local
			an_ele_name: NXML_NAME
		do
			-- inherit parent's context if none declared in this 
			-- start tag, otherwise use the one just built.
			if start_tag_context = Void then
				namespaces.force (namespaces.item)
			else
				namespaces.force (start_tag_context)
			end
			
			-- now the namespaces are ready, resolve element 
			-- name and queued attributes
			if not is_valid_name (start_tag_name) then
				force_error (is_valid_error (start_tag_name))
			else
				an_ele_name := new_name(start_tag_name)
				when_start (an_ele_name)
				process_attributes (an_ele_name.namespace)
				when_content_start
			end
		end

	when_end_tag (a_name: NXML_PARSER_NAME) is
			-- End tag.
		do
			when_end
			
			-- pop from namespace context stack
			namespaces.remove
		end

feature -- Actions

	when_comment (a_content: STRING) is
			-- Comment ignored.
		do
		end

	when_processing_instruction (a_target, a_content: STRING) is
			-- PI ignored.
		do
		end

feature {NONE} -- Context for start tag

	make_start_tag (a_name: NXML_PARSER_NAME) is
			-- Initialise start tag context.
		require
			not_void: a_name /= Void
		do
			start_tag_context := Void
			start_tag_name := a_name
			make_attribute_queue
		end
		
	start_tag_context: NXML_NAMESPACE_CONTEXT
			-- Namespace context build if there are NS declaration 
			-- in the tag.
			
	start_tag_name: NXML_PARSER_NAME
			-- Unresolved start tag.
	
feature {NONE} -- Attribute queue

	make_attribute_queue is
			-- Make attributes queue.
		do
			!! attribute_names.make
			!! attribute_values.make
		ensure
			not_void: attribute_names /= Void and attribute_values /= Void
			count_equal: attribute_names.count = attribute_values.count
			empty: attribute_names.is_empty
		end
		
	queue_attribute (a_name: NXML_PARSER_NAME; a_value: STRING) is
			-- Add one attribute to queue.
		do
			attribute_names.force (a_name)
			attribute_values.force (a_value)
		ensure
			count_equal: attribute_names.count = attribute_values.count
		end
		
	process_attributes (an_element_ns: NXML_NAMESPACE) is
			-- Add attribute at top of queue into table, and
			-- remove item from queue.
		require
			not_void: an_element_ns /= Void
		local
			a_name: NXML_PARSER_NAME
		do
			from
			invariant
				count_equal: attribute_names.count = attribute_values.count
			until
				attribute_names.count = 0
			loop
				a_name := attribute_names.item
				if is_valid_name (a_name) then
					when_attribute (new_attribute_name (an_element_ns, a_name), 
							attribute_values.item)
					-- next
					attribute_names.remove
					attribute_values.remove
				else
					force_error (is_valid_error (a_name))
					-- exit loop
					attribute_names.wipe_out
					attribute_values.wipe_out
				end
			end
		ensure
			count_equal: attribute_names.count = attribute_values.count
			count_zero: attribute_names.is_empty
		end
		
	attribute_names: DS_LINKED_QUEUE [NXML_PARSER_NAME]
	attribute_values: DS_LINKED_QUEUE [STRING]
			-- DS_LINKED_QUEUE [NXML_PARSER_NAME+STRING]
				
feature {NONE} -- Namespace context stack

	namespaces: DS_LINKED_STACK [NXML_NAMESPACE_CONTEXT]
	
	make_namespaces is
			-- Initialise namespace context stack.
		local
			a_context: NXML_NAMESPACE_CONTEXT
		do
			!! a_context.make
			!! namespaces.make
			namespaces.force (a_context)
		ensure
			done: namespaces /= Void
			has_top_context: namespaces.count = 1
		end
		
	namespace_context: NXML_NAMESPACE_CONTEXT is
			-- Current namespace context.
		do
			Result := namespaces.item
		end

feature {NONE} -- Name

	is_valid_name (a_name: NXML_PARSER_NAME): BOOLEAN is
			-- Valid for building NXML_NAME?
		do
			Result := (a_name /= Void)
				and then
					(
						a_name.is_namespace_declaration
					or else
						a_name.is_simple_name
					or else
	 					a_name.is_namespace_name and then namespace_context.has (a_name.first)
					)
		end

	is_valid_error (a_name:NXML_PARSER_NAME): STRING is
			-- Error for an invalid name.
		require
			not_valid: not is_valid_name (a_name)
		do
			if a_name = Void or else not a_name.is_name then
				Result := Error_namespaces_name_misformed
			else
				Result := Error_namespaces_undeclared
			end
		end
		
	new_name (a_name: NXML_PARSER_NAME): NXML_NAME is
			-- XML name from namespace + raw name, resolve namespace.
		require
			valid: is_valid_name (a_name)
		do
			if a_name.count = 1 then
				!! Result.make (namespace_context.namespace, a_name.first)
			else
				!! Result.make (namespace_context.named (a_name.first), a_name.last)
			end
			
			Result := cached_name (Result)
		end
		
	new_attribute_name (an_ele_space: NXML_NAMESPACE; a_name: NXML_PARSER_NAME): NXML_NAME is
			-- Name of attribute uses element's namespace if 
			-- not explicit.
		require
			space: an_ele_space /= Void
			valid: is_valid_name (a_name)
		do
			if a_name.count = 1 then
				!! Result.make (an_ele_space, a_name.first)
			else
				!! Result.make (namespace_context.named (a_name.first), a_name.last)
			end

			Result := cached_name (Result)
		end
					
feature {NONE} -- All neame
		
	cached_name (a_name: NXML_NAME): NXML_NAME is
			-- Use name from cache if present to minimise objets in 
			-- output tree and allow fast object comparisons.
		do
			if all_names.has (a_name) then
				Result := all_names.item (a_name)
			else
				Result := a_name
				all_names.force_new (Result)
			end
		end
		
	all_names: DS_HASH_SET [NXML_NAME]
			-- Set of all known names that can thus be shared, 
			-- assuming they are immutable.
	
	make_all_names is
			-- Init names cache.
		do
			!! all_names.make_equal (0)
		end
		
end
